/************************************************************************
 *  COMPILE.C                                                           *
 *                      Der Compiler fr PINC                           *
 *                      =====================                           *
 *                                                                      *
 *  24.06.93    Bodo Wenzel     V1.0, Erstellung                        *
 ************************************************************************/

/* === Einfgedateien ================================================= */

#include "stdio.h"

#include "pinc.h"
#include "preproc.h"
#include "symbol.h"
#include "expressn.h"
#include "machine.h"

#define extern
#include "compile.h"

/* === Konstanten ===================================================== */

#define ST_NONE         0
#define ST_IF           1       /* ST_NONE + 1                          */
#define ST_SWITCH       2       /* ST_IF + 1                            */
#define ST_WHILE        3       /* ST_SWITCH + 1                        */
#define ST_DO           4       /* ST_WHILE + 1                         */
#define ST_FOR          5       /* ST_DO + 1                            */
#define ST_RETURN       6       /* ST_FOR + 1                           */
#define ST_BREAK        7       /* ST_RETURN + 1                        */
#define ST_CONT         8       /* ST_BREAK + 1                         */
#define ST_EXP          9       /* ST_CONT + 1                          */

#define LQ_SYM          0
#define LQ_STK          1       /* LQ_SYM+1                             */
#define LQ_LOOP         2       /* LQ_STK+1                             */
#define LQ_QUIT         3       /* LQ_LOOP+1                            */
#define LQSIZE          4       /* LQ_QUIT+1                            */

#define LQMAX           100     /* / 4 = Anzahl Eintrge                */

/* === Globale Variablen ============================================== */

int     lq[LQMAX],              /* loop-Stapel                          */
        *lqptr;                 /* Zeiger auf freien Eintrag            */

int     labnext,                /* nchste freie Labelnummer            */
        funcvstk,               /* Stackpointer auf Parameter           */
        fl_extern,              /* Flag, wenn externe Variable          */
        lastst;                 /* letztes ausgefhrtes Statement       */

/* === Funktionsprototypen ============================================ */

cc_gdecl(int type);
cc_ldecl(int type);

cc_func();
cc_getarg();
cc_decl();

cc_stmt();
cc_compound();
cc_if();
cc_switch();
cc_while();
cc_do();
cc_for();
cc_return();
cc_break();
cc_continue();

cc_jfalse(int lab);
cc_jtrue(int lab);

cc_lqadd(char *sptr,int stk,int loop,int quit);
cc_lqdel();
cc_lqread();

cc_sub();
cc_semi();

/* === Funktionen ===================================================== */

/* --- Compiler ------------------------------------------------------- */

compile()                       /* Der Compiler                         */
/*-----*/
  {
  cfunc=NULL;                   /* bisher keine Funktion                */
  lqptr=lq;                     /*        keine Schleifen               */
  labnext=0;                    /* Labelzhler vorbereiten              */
  fl_extern=0;                  /* Variablen nicht extern               */

  while (ineof==0)              /* bis Quellenende...                   */
    {
    if (matcha("extern"))
      {
      fl_extern=1;
      if (matcha("char"))
        cc_gdecl(S_TY_CHR);     /* "extern char"                        */
      else
        {
        if (matcha("int")==0)
          error("Fehlende Typangabe, \"int\" angenommen");
        cc_gdecl(S_TY_INT);     /* "extern int"                         */
        }
      cc_semi();
      fl_extern=0;
      }
    else if (matcha("char"))
      {
      cc_gdecl(S_TY_CHR);       /* global "char"                        */
      cc_semi();
      }
    else if (matcha("int"))
      {
      cc_gdecl(S_TY_INT);       /* global "int"                         */
      cc_semi();
      }
    else
      cc_func();                /* sonst mu es eine Funktion sein      */
    }
  }

/* --- Variablen deklarieren ------------------------------------------ */

cc_gdecl(int type)              /* Eine globale Variable merken         */
/*--------------*/
  {
  int   n,id;
  char  name[NAMESIZE];

  while (1)
    {
    n=1;
    if (matchx("*"))            /* Zeiger oder nicht ?                  */
      id=S_ID_PTR;
    else
      id=S_ID_VAR;
    if (symname(name)==0)       /* Namen prfen                         */
      ername();
    else if (symfind(name))
      ermulti(name);
    if (matchx("["))            /* Vektor abfragen                      */
      {
      if (id==S_ID_PTR)
        erpointer();
      n=cc_sub();
      if (n)
        id=S_ID_ARR;
      else
        id=S_ID_PTR;
      }
    if (fl_extern)
      symadd(name,id,type,S_OF_EXT);
    else
      symadd(name,id,type,n);   /* Namen usw. merken                    */
    if (matchx(",")==0)
      return;                   /* wenn kein Komma, Ende                */
    }
  }

cc_ldecl(int type)              /* Eine lokale Variable merken          */
/*--------------*/
  {
  int   n,id;
  char  name[NAMESIZE];

  while (1)
    {
    if (matchx("*"))            /* Zeiger oder nicht ?                  */
      id=S_ID_PTR;
    else
      id=S_ID_VAR;
    if (symname(name)==0)       /* Namen prfen                         */
      ername();
    else if (symfind(name)>=symlptr)
      ermulti(name);
    if (matchx("["))            /* Vektor abfragen                      */
      {
      if (id==S_ID_PTR)
        erpointer();
      n=cc_sub();
      if (n)
        {
        id=S_ID_ARR;
        if (type==S_TY_INT)     /* Gre anpassen                       */
          n=n*mintwidth;
        else
          n=((n-1)/mcharaln+1)*mcharaln;
        }
      else
        {
        id=S_ID_PTR;
        n=mintwidth;
        }
      }
    else
      if ((type==S_TY_CHR) & (id!=S_ID_PTR))
        n=mcharaln;
      else
        n=mintwidth;            /* Gre anpassen                       */
    if (mstackdir<0)
      {
      funcvstk=funcvstk-n;
      symadd(name,id,type,funcvstk);    /* Namen usw. merken            */
      }
    else
      {
      symadd(name,id,type,funcvstk);    /* Namen usw. merken            */
      funcvstk=funcvstk+n;
      }
    if (matchx(",")==0)
      return;                   /* wenn kein Komma, Ende                */
    }
  }

/* === Funktionsdefinition ============================================ */

cc_func()                       /* Funktion compilieren                 */
/*-----*/
  {
  char  name[NAMESIZE];
  int   *iptr,stmt;

  if (ineof)                    /* Quellenende abfangen                 */
    return;
  funcline=1;                   /* erste Zeile                          */

  if (symname(name)==0)         /* Namen prfen                         */
    {
    ername();
    sprintf(name,"cc%d",++labnext);
    }
  if (cfunc=symfind(name))      /* Name schon bekannt ?                 */
    {
    iptr=cfunc+SYM_OFFSET;
    if (cfunc[SYM_ID]!=S_ID_FNC)
      ermulti(name);
    }
  else
    {
    cfunc=symadd(name,S_ID_FNC,S_TY_INT,S_OF_EXT);
    iptr=cfunc+SYM_OFFSET;
    }
  symgptr=symlptr=symptr;       /* Symbolzeiger merken                  */
  funclstk=0;

  ccneedchar("(");
  funcvstk=mretwidth;
  if (matchx(")")==0)           /* Argumente abfragen                   */
    {
    cc_getarg();                /* arbeitet rekursiv !!!                */
    ccneedchar(")");
    }

  if (matchx(";")==0)           /* kein Prototyp ?                      */
    {
    if (cfunc[SYM_ID]==S_ID_FNC &
        *iptr!=S_OF_EXT & *iptr!=S_OF_REF)
      ermulti(name);
    *iptr=S_OF_FNC;             /* Funktion jetzt definiert             */

    if (verbose)
      {
      sprintf(oline,"%s()\r\n",name);
      puts(oline);
      }
    moutcmdn(T_PFUNC,name);     /* Funktionskopf                        */

    stmt=ST_NONE;
    ccneedchar("{");
    cc_decl();                  /* lokale Variablen aufnehmen           */
    while (matchx("}")==0)
      {
      stmt=cc_stmt();           /* Statements bersetzen                */
      if (ineof)
        {
        error("Unerwartetes Ende der Quelldatei");
        break;
        }
      }
    if (stmt!=ST_RETURN)        /* bei Bedarf "return" compilieren      */
      {
      mmodifysp(0);
      moutcmd(T_RET);
      }
    }
  symptr=symgptr;               /* lokale Namen vergessen               */
  cfunc=NULL;
  }

cc_getarg()                     /* Ein Argument anlegen (rekursiv)      */
/*-------*/
  {
  char  name[NAMESIZE],*s;
  int   *iptr,id,type;

  if (matcha("char"))           /* Typ abfragen                         */
    type=S_TY_CHR;
  else
    {
    type=S_TY_INT;
    if (matcha("int")==0)
      error("Falscher Typ, \"int\" angenommen");
    }
  if (matchx("*"))              /* Zeiger oder nicht ?                  */
    id=S_ID_PTR;
  else
    id=S_ID_VAR;
  if (symname(name)==0)         /* Namen prfen                         */
    ername();
  if ((s=symfind(name))>=symlptr)
    ermulti(name);
  else
    s=symadd(name,id,type,S_OF_NONE);
  if (matchx("["))              /* Vektor abfragen                      */
    {
    if (id==S_ID_PTR)
      erpointer();
    cc_sub();
    s[SYM_ID]=id=S_ID_PTR;
    }

  if (matchx(","))              /* bei Bedarf nchstes Argument         */
    cc_getarg();

  iptr=s+SYM_OFFSET;            /* Stackoffset mitfhren                */
  if (mstackdir<0)
    {
    if ((type==S_TY_CHR) & (id==S_ID_VAR))
      *iptr=funcvstk+mcharoff;
    else
      *iptr=funcvstk;
    funcvstk=funcvstk+mintwidth;
    }
  else
    {
    funcvstk=funcvstk+mintwidth;
    if ((type==S_TY_CHR) & (id==S_ID_VAR))
      *iptr=-funcvstk+mcharoff;
    else
      *iptr=-funcvstk;
    }
  }

/* === Statements ===================================================== */

cc_decl()                       /* Lokale Variablen anlegen             */
/*-----*/
  {
  funcvstk=funclstk;            /* aktuellen Stackpointer merken        */
  while (1)
    {
    if (ineof)
      break;
    else if (matcha("char"))
      {
      cc_ldecl(S_TY_CHR);       /* lokal "char"                         */
      cc_semi();
      }
    else if (matcha("int"))
      {
      cc_ldecl(S_TY_INT);       /* lokal "int"                          */
      cc_semi();
      }
    else
      break;
    }
  funclstk=mmodifysp(funcvstk); /* Stackpointer setzen                  */
  }

cc_stmt()                       /* Ein statement compilieren            */
/*-----*/
  {
  if (ineof)
    return(ST_NONE);
  else if (matchx("{"))         /* Block                                */
    cc_compound();
  else if (matcha("if"))        /* "if"                                 */
    {
    cc_if();
    lastst=ST_IF;
    }
  else if (matcha("switch"))    /* "switch"                             */
    {
    cc_switch();
    lastst=ST_SWITCH;
    }
  else if (matcha("while"))     /* "while"                              */
    {
    cc_while();
    lastst=ST_WHILE;
    }
  else if (matcha("do"))        /* "do-while"                           */
    {
    cc_do();
    lastst=ST_DO;
    }
  else if (matcha("for"))       /* "for"                                */
    {
    cc_for();
    lastst=ST_FOR;
    }
  else if (matcha("return"))    /* "return"                             */
    {
    cc_return();
    lastst=ST_RETURN;
    }
  else if (matcha("break"))     /* "break"                              */
    {
    cc_break();
    cc_semi();
    lastst=ST_BREAK;
    }
  else if (matcha("continue"))  /* "continue"                           */
    {
    cc_continue();
    cc_semi();
    lastst=ST_CONT;
    }
  else if (matchx(";"))         /* leere Anweisung                      */
    ;
  else                          /* Ausdruck                             */
    {
    expression();
    cc_semi();
    lastst=ST_EXP;
    }
  return(lastst);
  }

/* --- Block ---------------------------------------------------------- */

cc_compound()
/*---------*/
  {
  int   stksave;
  char  *symsave;

  symsave=symlptr;
  symlptr=symptr;
  stksave=funclstk;             /* momentane Zeiger merken              */

  cc_decl();                    /* lokale Variablen aufnehmen           */
  while (matchx("}")==0)
    {
    cc_stmt();                  /* Statements bersetzen                */
    if (ineof)
      break;
    }
  funclstk=mmodifysp(stksave);
  symptr=symlptr;
  symlptr=symsave;              /* Zeiger zurcksetzen                  */
  }

/* --- IF ------------------------------------------------------------- */

cc_if()
/*---*/
  {
  int   lab1,lab2;

  lab1=++labnext;
  cc_jfalse(lab1);              /* Vergleich                            */
  cc_stmt();                    /* positives Statement                  */

  if (matcha("else")==0)
    {
    moutcmdl(T_LABEL,lab1);
    return;
    }
  moutcmdl(T_JUMP,lab2=++labnext);
  moutcmdl(T_LABEL,lab1);
  cc_stmt();                    /* alternatives Statement               */
  moutcmdl(T_LABEL,lab2);
  }

/* --- SWITCH --------------------------------------------------------- */

cc_switch()
/*-------*/
  {
  int   lab1,lab2,num,stmt;

  lab1=0;
  lab2=0;
  cc_lqadd(symptr,funclstk,0,++labnext);        /* Schleife merken      */

  ccneedchar("(");
  expression();
  ccneedchar(")");
  mpush();                      /* Verzweigungsausdruck bersetzen      */

  ccneedchar("{");
  stmt=ST_NONE;
  while (matchx("}")==0)
    {
    if (ineof)
      break;
    if (matcha("case"))         /* "case"                               */
      {
      if ((stmt!=ST_BREAK) & (lab2!=0)) /* evtl. Test berspringen      */
        moutcmdl(T_JUMP,lab2);
      if (lab1>0)               /* bei Bedarf Marke fr diesen Test     */
        moutcmdl(T_LABEL,lab1);
      lab1=++labnext;

      if (enumber(&num)==0)
        if (echar(&num)==0)
          {
          error("Fehlerhafte Konstante");
          num=0;
          skipjunk();
          }
      moutcmdv(T_CPIMM,num);
      ccneedchar(":");
      moutcmdl(T_NEJMP,lab1);   /* Test bersetzen                      */

      if ((stmt!=ST_BREAK) & (lab2!=0)) /* evtl. Marke fr berspringen */
        moutcmdl(T_LABEL,lab2);
      lab2=++labnext;
      stmt=ST_NONE;
      }
    else if (matcha("default")) /* "default"                            */
      {
      if (lab1>0)               /* bei Bedarf Marke fr "default"       */
        moutcmdl(T_LABEL,lab1);
      ccneedchar(":");
      lab1=-1;
      lab2=++labnext;
      stmt=ST_NONE;
      }
    else
      {
      if (lab1==0)
        {
        error("Fehlendes \"case\"");
        lab1=-1;
        lab2=++labnext;
        }
      stmt=cc_stmt();           /* Statements bersetzen                */
      }
    }
  if (lab1>0)
    moutcmdl(T_LABEL,lab1);     /* evtl. Marke fr berspringen         */

  cc_lqdel();                   /* Schleife vergessen                   */
  }

/* --- WHILE ---------------------------------------------------------- */

cc_while()
/*------*/
  {
  int   lab1,lab2;

  lab1=++labnext;
  lab2=++labnext;
  cc_lqadd(symptr,funclstk,lab1,lab2);  /* Schleife merken              */

  moutcmdl(T_LABEL,lab1);
  cc_jfalse(lab2);              /* Bedingung bersetzen                 */
  cc_stmt();                    /* Statement bersetzen                 */
  moutcmdl(T_JUMP,lab1);        /* Marke fr berspringen               */

  cc_lqdel();                   /* Schleife vergessen                   */
  }

/* --- DO ------------------------------------------------------------- */

cc_do()
/*---*/
  {
  int   lab1,lab2;

  lab1=++labnext;
  lab2=++labnext;
  cc_lqadd(symptr,funclstk,lab2,++labnext);     /* Schleife merken      */

  moutcmdl(T_LABEL,lab1);
  cc_stmt();                    /* Statement bersetzen                 */
  moutcmdl(T_LABEL,lab2);
  if (matcha("while")==0)
    error("Fehlendes \"while\"");
  cc_jtrue(lab1);               /* Bedingung bersetzen                 */
  cc_semi();

  cc_lqdel();                   /* Schleife vergessen                   */
  }

/* --- FOR ------------------------------------------------------------ */

cc_for()
/*----*/
  {
  int   lab1,lab2,lab3,lab4;

  lab1=++labnext;
  lab2=++labnext;
  lab3=++labnext;
  lab4=++labnext;
  cc_lqadd(symptr,funclstk,lab3,lab4);  /* Schleife merken              */

  ccneedchar("(");
  if (matchx(";")==0)
    {
    expression();               /* bei Bedarf Initialisierungsausdruck  */
    cc_semi();
    }
  moutcmdl(T_LABEL,lab1);
  if (matchx(";")==0)
    {
    expression();               /* bei Bedarf Bedingung                 */
    cc_semi();
    moutcmdl(T_FJUMP,lab4);
    }
  moutcmdl(T_JUMP,lab2);        /* Sprung zum Schleifenkrper           */

  moutcmdl(T_LABEL,lab3);
  if (matchx(")")==0)
    {
    expression();               /* bei Bedarf Schleifenausdruck         */
    ccneedchar(")");
    }
  moutcmdl(T_JUMP,lab1);
  moutcmdl(T_LABEL,lab2);
  cc_stmt();                    /* Statement bersetzen                 */
  moutcmdl(T_JUMP,lab3);

  cc_lqdel();                   /* Schleife vergessen                   */
  }

/* --- RETURN --------------------------------------------------------- */

cc_return()
/*-------*/
  {
  if (matchx(";")==0)
    {
    expression();               /* bei Bedarf Ausdruck bersetzen       */
    cc_semi();
    }
  mmodifysp(0);
  moutcmd(T_RET);
  }

/* --- BREAK ---------------------------------------------------------- */

cc_break()
/*------*/
  {
  int   *ptr;

  if ((ptr=cc_lqread())==0)
    return;
  mmodifysp(ptr[LQ_STK]);       /* zum Schleifenende                    */
  moutcmdl(T_JUMP,ptr[LQ_QUIT]);
  }

/* --- CONTINUE ------------------------------------------------------- */

cc_continue()
/*---------*/
  {
  int   *ptr;

  if ((ptr=cc_lqread())==0)
    return;
  if (ptr[LQ_LOOP]==0)          /* in "switch" ignorieren               */
    return;
  mmodifysp(ptr[LQ_STK]);       /* zur Schleifenwiederholung            */
  moutcmdl(T_JUMP,ptr[LQ_LOOP]);
  }

/* --- Funktionsaufrufe und Sprnge ----------------------------------- */

cccall(char *sptr)              /* Funktionsaufruf                      */
/*--------------*/
  {
  int   stksave;

  stksave=funclstk;
  if (sptr==0)                  /* bei Bedarf Adresse merken            */
    mpush();
  while (ppactch()!=')')
    {
    if (ineof)
      break;
    expression();               /* Parameter als Ausdruck bersetzen    */
    mpush();
    if (matchx(",")==0)
      break;
    }
  ccneedchar(")");
  if (sptr)                     /* je nach Art Aufruf bersetzen        */
    moutcmdn(T_CALL,sptr+SYM_NAME);
  else
    if (mstackdir<0)
      moutcmdv(T_CALLS,stksave-funclstk-mintwidth);
    else
      moutcmdv(T_CALLS,stksave-funclstk);
  funclstk=mmodifysp(stksave);  /* Stack wieder zurcksetzen            */
  }

cc_jfalse(int lab)              /* Sprung, wenn falsch                  */
/*--------------*/
  {
  ccneedchar("(");
  expression();                 /* Ausdruck bersetzen                  */
  ccneedchar(")");
  moutcmdl(T_FJUMP,lab);
  }

cc_jtrue(int lab)               /* Sprung, wenn wahr                    */
/*-------------*/
  {
  ccneedchar("(");
  expression();                 /* Ausdruck bersetzen                  */
  ccneedchar(")");
  moutcmdl(T_TJUMP,lab);
  }

/* --- Funktionen zur Schleifenkontrolle ------------------------------ */

cc_lqadd(char *sptr,int stk,int loop,int quit)  /* Neue Schleife        */
/*------------------------------------------*/
  {
  if (lqptr>=&lq[LQMAX])
    {
    error("Zu tief geschachtelte Schleifen");
    return;
    }
  lqptr[LQ_SYM]=sptr;           /* diverse Zeiger merken                */
  lqptr[LQ_STK]=stk;
  lqptr[LQ_LOOP]=loop;
  lqptr[LQ_QUIT]=quit;
  lqptr=lqptr+LQSIZE;           /* nchster Eintrag                     */
  }

cc_lqdel()                      /* Innerste Schleife beenden            */
/*------*/
  {
  if (cc_lqread())
    {
    lqptr=lqptr-LQSIZE;
    symlptr=symptr=lqptr[LQ_SYM];
    funclstk=mmodifysp(lqptr[LQ_STK]);
    moutcmdl(T_LABEL,lqptr[LQ_QUIT]);   /* Zeiger zurcksetzen          */
    }
  }

cc_lqread()                     /* Innerste Schleife holen              */
/*-------*/
  {
  if (lqptr==lq)
    {
    error("Keine aktiven Schleifen");
    return(NULL);
    }
  else
    return(lqptr-LQSIZE);
  }

/* --- Abfrage nach syntaktisch notwendigen Tokens -------------------- */

cc_sub()                        /* Feldgre abfragen                   */
/*----*/
  {
  int   num;

  if (matchx("]"))              /* keine Angabe = Zeiger                */
    return(0);
  if (enumber(&num)==0)
    {
    skipjunk();
    error("Feldgre mu Konstante sein");
    num=1;
    }
  else if (num<=0)
    {
    error("Fehlerhafte Feldgre");
    num=1;
    }
  ccneedchar("]");
  return(num);
  }

ccneedchar(char *c)             /* Zeichen fordern                      */
/*---------------*/
  {
  if (matchx(c)==0)
    {
    sprintf(oline,"Fehlendes Zeichen: '%s'",c);
    error(oline);
    }
  }

cc_semi()                       /* Semikolon fordern                    */
/*-----*/
  {
  if (matchx(";")==0)
    error("Fehlendes Semikolon");
  }

/* === Ende =========================================================== */
