/************************************************************************
 *  MACHINE.C                                                           *
 *                      Codeausgabe fr PINC                            *
 *                      ====================                            *
 *                                                                      *
 *  22.08.93    Bodo Wenzel     V1.0, Erstellung                        *
 ************************************************************************/

/* === Einfgedateien ================================================= */

#include "ctype.h"
#include "stdio.h"
#include "string.h"

#include "pinc.h"
#include "compile.h"
#include "symbol.h"
#include "expressn.h"

#define extern
#include "machine.h"

/* === Konstanten ===================================================== */

#define M_EXTBEG        0x01
#define M_EXTEND        0x02    /* Codes fr extern-Symbol              */

#define T_APREF         1       /* Namensprefix                         */
#define T_ASUFF         2       /* Namenssuffix                         */
#define T_MVAL          3       /* maximale Anzahl Werte pro Zeile      */
#define T_SDIR          4       /* Stackrichtung                        */
#define T_RWID          5       /* Anzahl Bytes je "return"             */
#define T_IWID          6       /* Anzahl Bytes je "int"                */
#define T_COFFS         7       /* Offset fr "char" in "int"           */
#define T_CALGN         8       /* Bytes Aufrundung fr "char"          */
#define TABVARS         9

#define TABSIZE         6000    /* Gre der bersetzungstabelle        */

/* === Globale Variablen ============================================== */

char    mlit[TABSIZE];          /* Puffer fr Tabelle                   */
int     mtab[TABMAX];           /* Zeiger auf die Eintrge              */

int     masmpref,               /* Anzahl erste Zeichen fr Symbole     */
        masmsuff,               /* Anzahl letzte Zeichen fr Symbole    */
        mmaxval;                /* Maximale Anzahl Bytes pro Zeile      */

/* === Funktionsprototypen ============================================ */

m_outtext(char *s);             /* Text ausgeben                        */
m_outchar(char c);              /* Zeichen ausgeben                     */

/* === Funktionen ===================================================== */

/* --- Tabelle laden -------------------------------------------------- */

mreadtab(char *name)            /* Tabelle lesen                        */
/*----------------*/
  {
  int   vars,file,i,j,in;
  char  *s,n[NAMESIZE];

  masmpref=-1;
  masmsuff=-1;
  mmaxval=0;
  mstackdir=0;
  mretwidth=-1;
  mintwidth=-1;
  mcharoff=-1;
  mcharaln=-1;                  /* Variablen vorbesetzen                */
  vars=0;
  if ((file=fopen(name,"r"))!=NULL)
    {
    in=0;                       /* bisher nicht in Feld                 */
    s=mlit;
    i=0;                        /* Zeiger setzen                        */
    while (1)
      {
      if (fgets(iline,ILINESIZE,file)==NULL)
        break;
      if (iline[0]==0x1A)       /* ^Z als EOF behandeln                 */
        break;
      ilineptr=iline;
      if (iline[0]=='#')
        {
        if (iline[1]=='#')      /* "##" zu "#"                          */
          ++ilineptr;
        else if (iline[1]=='>') /* "#>" als "extern"-Deklaration        */
          {
          iline[1]=M_EXTBEG;
          ilineptr=iline+1;
          while (isspace(*++ilineptr));
          if (symname(n)==0)
            errfatal("Fehlerhaftes Symbol in Tabelle",RET_TABLE);
          symadd(n,S_ID_FNC,S_TY_INT,S_OF_EXT);
          ilineptr=iline+1;
          j=-1;
          while (*++ilineptr=n[++j]);
          *ilineptr++=M_EXTEND;
          *ilineptr='\0';       /* Namen bernehmen                     */
          ilineptr=iline+1;
          }
        else                    /* sonst als Feldtrenner                */
          {
          in=0;                 /* nicht mehr in Feld                   */
          continue;
          }
        }
      if (in)
        --s;                    /* nchste Zeile im Feld anschlieen    */
      else
        {
        if ((i==TABVARS) & (vars==0))   /* Variablen gesammelt ?        */
          {
          sscanf(mtab[T_APREF],"%d",&masmpref);
          sscanf(mtab[T_ASUFF],"%d",&masmsuff);
          sscanf(mtab[T_MVAL],"%d",&mmaxval);
          sscanf(mtab[T_SDIR],"%d",&mstackdir);
          sscanf(mtab[T_RWID],"%d",&mretwidth);
          sscanf(mtab[T_IWID],"%d",&mintwidth);
          sscanf(mtab[T_COFFS],"%d",&mcharoff);
          sscanf(mtab[T_CALGN],"%d",&mcharaln);
          if (masmpref<=-1 |
              masmsuff<=-1 |
              masmpref+masmsuff==0 |
              mmaxval<=0 |
              mstackdir==0 |
              mretwidth<=-1 |
              mintwidth<=-1 |
              mcharoff<=-1 |
              mcharaln<=-1)
            errfatal("Falsche Zahlenwerte in Tabelle",RET_TABLE);
          i=i-TABVARS+1;
          s=mtab[i];
          vars=1;               /* Zeiger zurcksetzen                  */
          }
        else if (i==TABMAX)
          errfatal("Zuviele Eintrge in Tabelle",RET_TABLE);

        mtab[i++]=s;            /* Feldanfang merken                    */
        }
      in=1;
      while (*s=*ilineptr++)    /* Zeile speichern                      */
        if (++s>=mlit+TABSIZE)
          errfatal("Tabelle zu gro",RET_TABLE);
      ++s;
      }
    fclose(file);
    if (i<TABMAX)
      errfatal("Zuwenig Eintrge in Tabelle",RET_TABLE);
    }
  else
    errfatal("Kann Tabelle nicht laden",RET_TABLE);
  }

/* --- Textausgabe ---------------------------------------------------- */

mheader(char *arg)              /* Kopf ausgeben                        */
/*--------------*/
  {
  moutcmds(T_COMM,title);
  moutcmds(T_COMM,arg);
  moutcmds(T_COMM,mtab[T_NAME]);
  moutcmds(T_COMM," --- Anfang ---");
  }

mextfuncs()                     /* Externe Funktionsnamen ausgeben      */
/*-------*/
  {
  char  *s;
  int   *iptr;

  s=symbeg;
  while (s<symptr)
    {
    iptr=s+SYM_OFFSET;
    if (s[SYM_ID]==S_ID_FNC & *iptr==S_OF_REF)
      moutcmdn(T_XFUNC,s+SYM_NAME);
    s=s+SYMSIZE;
    }
  }

mdataout()                      /* Konstanten ausgeben                  */
/*------*/
  {
  int   j,k;
  char  lline[OLINESIZE],*ll;

  moutcmdl(T_CLBL,0);
  for (k=0; k<litptr; )
    {
    j=mmaxval;
    ll=lline;
    while (j--)
      {
      sprintf(ll,"%d",lit[k++]);
      ll=ll+strlen(ll);
      if (j==0 | k>=litptr)
        break;
      *ll++=',';
      }
    moutcmds(T_DCHR,lline);
    }
  }

mbssout()                       /* Globale Variablen ausgeben           */
/*-----*/
  {
  char  *s;
  int   *iptr,n;

  s=symbeg;
  while (s<symptr)
    {
    if (s[SYM_ID]!=S_ID_FNC)
      {
      iptr=s+SYM_OFFSET;
      if (*iptr==S_OF_REF)
        {
        if (s[SYM_TYPE]==S_TY_INT | s[SYM_ID]==S_ID_PTR)
          moutcmdn(T_XINT,s+SYM_NAME);
        else
          moutcmdn(T_XCHR,s+SYM_NAME);
        }
      else if (*iptr!=S_OF_EXT)
        {
        if (s[SYM_TYPE]==S_TY_INT | s[SYM_ID]==S_ID_PTR)
          {
          moutcmdn(T_PINT,s+SYM_NAME);
          moutcmdv(T_DIST,*iptr);
          }
        else
          {
          n=((*iptr-1)/mcharaln+1)*mcharaln;    /* Gre anpassen       */
          moutcmdn(T_PCHR,s+SYM_NAME);
          moutcmdv(T_DCST,n);
          }
        }
      }
    s=s+SYMSIZE;
    }
  }

mtrailer()                      /* Ende ausgeben                        */
/*------*/
  {
  moutcmd(T_END);
  moutcmds(T_COMM," --- Ende ---");
  }

/* --- Speicherzugriffe ----------------------------------------------- */

mgetgbl(char *s)                /* PRI = globale Variable               */
/*------------*/
  {
  if (s[SYM_ID]!=S_ID_PTR & s[SYM_TYPE]==S_TY_CHR)
    moutcmdn(T_LACHR,s+SYM_NAME);
  else
    moutcmdn(T_LAINT,s+SYM_NAME);
  }

mputgbl(char *s)                /* Globale Variable = PRI               */
/*------------*/
  {
  if (s[SYM_ID]!=S_ID_PTR & s[SYM_TYPE]==S_TY_CHR)
    moutcmdn(T_SACHR,s+SYM_NAME);
  else
    moutcmdn(T_SAINT,s+SYM_NAME);
  }

mgetgadr(char *s)               /* PRI = globale Adresse                */
/*-------------*/
  {
  moutcmdn(T_LGADR,s+SYM_NAME);
  }

mgetladr(char *s)               /* PRI = lokale Adresse                 */
/*-------------*/
  {
  int   *iptr;

  iptr=s+SYM_OFFSET;
  moutcmdv(T_LLADR,*iptr-funclstk);
  }

mgetdadr(int offs)              /* PRI = Konstantenadresse              */
/*--------------*/
  {
  char  b[20];

  sprintf(b,"cc0+%d",offs);
  moutcmds(T_LDADR,b);
  }

mgetind(char type)              /* PRI = *PRI                           */
/*--------------*/
  {
  if (type==S_TY_CHR)
    moutcmd(T_LICHR);
  else
    moutcmd(T_LIINT);
  }

mputind(char type)              /* *TOS = PRI                           */
/*--------------*/
  {
  mpop();
  if (type==S_TY_INT)
    moutcmd(T_SIINT);
  else
    moutcmd(T_SICHR);
  }

/* --- Stack- und Registeroperationen --------------------------------- */

mpush()                         /* PRI in Stack                         */
/*---*/
  {
  moutcmd(T_PUSH);
  if (mstackdir<0)
    funclstk=funclstk-mintwidth;
  else
    funclstk=funclstk+mintwidth;
  }

mpop()                          /* Ein Wert vom Stack                   */
/*--*/
  {
  if (mstackdir<0)
    funclstk=funclstk+mintwidth;
  else
    funclstk=funclstk-mintwidth;
  }

mmodifysp(int newsp)            /* Stackpointer verndern               */
/*----------------*/
  {
  int   k;

  k=newsp-funclstk;
  if (k)
    moutcmdv(T_CORSP,k);
  return(newsp);
  }

/* --- Ausgabefunktionen ---------------------------------------------- */

moutcmd(int cmd)                /* Code ausgeben                        */
/*------------*/
  {
  m_outtext(mtab[cmd]);
  }

moutcmdn(int cmd,char *n)       /* Code und Symbolnamen ausgeben        */
/*---------------------*/
  {
  char  b[CNAMESIZE],*s;
  int   len,i;

  s=b;
  len=strlen(n);
  i=-1;
  while (++i<len)
    if (islower(n[i]))
      {
      *s++='c';
      *s++='c';
      break;
      }
  if (len > masmpref+masmsuff)
    {
    i=masmpref;
    len=len-masmpref-masmsuff;
    while (i-- > 0)
      *s++=*n++;
    while (len-- > 0)
      n++;
    }
  while ((*s++=*n++)!='\0');
  sprintf(oline,mtab[cmd],b);
  m_outtext(oline);
  }

moutcmdl(int cmd,int l)         /* Code und Sprungmarke ausgeben        */
/*-------------------*/
  {
  char  b[CNAMESIZE];

  sprintf(b,"cc%d",l);
  sprintf(oline,mtab[cmd],b);
  m_outtext(oline);
  }

moutcmdv(int cmd,int v)         /* Code und Wert ausgeben               */
/*-------------------*/
  {
  sprintf(oline,mtab[cmd],v);
  m_outtext(oline);
  }

moutcmds(int cmd,char *s)       /* Code und Text ausgeben               */
/*---------------------*/
  {
  sprintf(oline,mtab[cmd],s);
  m_outtext(oline);
  }

m_outtext(char *s)              /* Text ausgeben                        */
/*--------------*/
  {
  char   *e;
  int    *iptr;

  while (*s)
    if (*s==M_EXTBEG)           /* externes Symbol in den Ausgabetext   */
      {
      e=++s;
      while (*++e!=M_EXTEND);
      *e='\0';
      iptr=symfind(s)+SYM_OFFSET;
      *iptr=S_OF_REF;
      *e=M_EXTEND;
      s=e+1;
      }
    else
      m_outchar(*s++);
  }

moutstr(char *s)                /* Zeichenkette ausgeben                */
/*------------*/
  {
  while (m_outchar(*s++));
  }

mouteol()                       /* Zeilenende ausgeben                  */
/*-----*/
  {
  m_outchar('\r');
  m_outchar('\n');
  }

/* --- Ausgabedatei --------------------------------------------------- */

m_outchar(char c)               /* Zeichen ausgeben                     */
/*-------------*/
  {
  if (c=='\0')
    return('\0');
  if (outfile!=NULL)
    if (fputc(c,outfile)==EOF)
      errfatal("Schreibfehler auf Ausgabedatei",RET_WRITE);
  return(c);
  }

/* === Ende =========================================================== */
